package probe

import (
	"context"
	"fmt"
	"net/http"
	"path"
	"sync"
	"time"

	"git.loafle.net/commons/configuration-go"
	cdr "git.loafle.net/commons/di-go/registry"
	logging "git.loafle.net/commons/logging-go"
	crc "git.loafle.net/commons/rpc-go/client"
	crpj "git.loafle.net/commons/rpc-go/protocol/json"
	crr "git.loafle.net/commons/rpc-go/registry"
	csswc "git.loafle.net/commons/server-go/socket/web/client"
	occa "git.loafle.net/overflow/commons-go/core/annotation"
	ocnc "git.loafle.net/overflow/commons-go/noauthprobe/config"
	ocncc "git.loafle.net/overflow/commons-go/noauthprobe/constants"
	ocpc "git.loafle.net/overflow/commons-go/probe/config"
	"git.loafle.net/overflow/probe/auth/info"
	_ "git.loafle.net/overflow/probe/auth/service"
	"git.loafle.net/overflow/probe/config"
)

type Probe struct {
	Config    *config.Config
	ConfigDir string

	client *crc.Client

	stopChan chan struct{}
	stopWg   sync.WaitGroup
}

func (p *Probe) Start() error {
	if p.stopChan != nil {
		return fmt.Errorf("already running. Stop it before starting it again")
	}

	if err := p.initClient(); nil != err {
		return err
	}

	if err := p.client.Start(); nil != err {
		return err
	}

	p.stopChan = make(chan struct{})

	p.stopWg.Add(1)
	go p.handleProbe()

	return nil
}

func (p *Probe) Stop(ctx context.Context) error {
	if p.stopChan == nil {
		return fmt.Errorf("must be started before stopping it")
	}
	close(p.stopChan)
	p.stopWg.Wait()

	p.stopChan = nil

	return nil
}

func (p *Probe) logHeader() string {
	return "Probe:"
}

func (p *Probe) initService() (crr.RPCInvoker, error) {
	services, err := cdr.GetInstancesByAnnotationType(occa.RPCServiceAnnotationType)
	if nil != err {
		return nil, err
	}

	rpcRegistry := crr.NewRPCRegistry()
	rpcRegistry.RegisterServices(services...)

	return rpcRegistry, nil
}

func (p *Probe) initClient() error {
	_connector := p.Config.Central.Connector.Clone()
	connector := _connector.(*csswc.Connectors)

	header := make(map[string][]string)

	switch p.authConfig.State() {
	case ocnc.AuthStateTypeRegisterd:
		header[ocncc.HTTPRequestHeaderKey_NoAuthProbe_Method] = []string{ocncc.HTTPRequestHeaderValue_NoAuthProbe_Method_Connect}
		header[ocncc.HTTPRequestHeaderKey_NoAuthProbe_TempProbeKey] = []string{*p.authConfig.TempKey}
	default:
		rh, err := info.GetRegistHeader(p.Config.Account.APIKey)
		if nil != err {
			return err
		}
		header[ocncc.HTTPRequestHeaderKey_NoAuthProbe_Method] = []string{ocncc.HTTPRequestHeaderValue_NoAuthProbe_Method_Regist}
		header[ocncc.HTTPRequestHeaderKey_NoAuthProbe_Info] = []string{rh}
	}

	centralURL := fmt.Sprintf("ws://%s:%d%s", p.Config.Central.Host, p.Config.Central.Port, ocncc.HTTPEntry_Auth)

	logging.Logger().Debugf("%s central %s", p.logHeader(), centralURL)

	connector.URL = centralURL
	connector.RequestHeader = header
	connector.ResponseHandler = func(res *http.Response) {
		switch p.authConfig.State() {
		case ocnc.AuthStateTypeNotRegisterd:
			tempProbeKey := res.Header.Get(ocncc.HTTPResponseHeaderKey_NoAuthProbe_SetTempProbeKey)
			p.tempKeyChan <- tempProbeKey
		default:
		}
	}

	rpcInvoker, err := p.initService()
	if nil != err {
		return err
	}
	codec := crpj.NewClientCodec()

	p.client = &crc.Client{
		Connector:  connector,
		Codec:      codec,
		RPCInvoker: rpcInvoker,
		Name:       "Authenticator",
	}

	return nil
}

func (p *Probe) handleProbe(endChan chan<- error) {
	var err error
	defer func() {
		if nil != p.client {
			err = p.client.Stop(context.Background())
		}

		p.stopWg.Done()
		endChan <- err
	}()

	for {
		select {
		case probeKey, ok := <-p.accpetedChan:
			if !ok {
				return
			}
			logging.Logger().Infof("%s accepted by central", p.logHeader())
			p.Config.Probe.Key = &probeKey
			err = configuration.Save(p.Config, path.Join(p.ConfigDir, ocpc.ConfigFileName), true)
			if nil != err {
				logging.Logger().Errorf("%s %v", p.logHeader(), err)
			}
			return
		case _, ok := <-p.deniedChan:
			if !ok {
				return
			}
			logging.Logger().Infof("%s denied by central", p.logHeader())
			n := time.Now()
			p.authConfig.DenyDate = &n
			err = configuration.Save(p.authConfig, path.Join(p.ConfigDir, ocnc.ConfigFileName), true)
			if nil != err {
				logging.Logger().Errorf("%s %v", p.logHeader(), err)
			}
			return
		case tempKey, ok := <-p.tempKeyChan:
			if !ok {
				return
			}
			logging.Logger().Infof("%s registered by central", p.logHeader())
			p.authConfig.TempKey = &tempKey
			err = configuration.Save(p.authConfig, path.Join(p.ConfigDir, ocnc.ConfigFileName), true)
			if nil != err {
				logging.Logger().Errorf("%s %v", p.logHeader(), err)
				return
			}
		case <-p.stopChan:
			return
		}
	}
}
